# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot
from uproot.interpretation.jagged import AsJagged
from uproot.interpretation.numerical import AsDtype


def test_histograms_outside_of_ttrees():
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        contents = numpy.asarray(f["hpx"].bases[-1])
        assert (contents.min(), contents.max()) == (0.0, 2417.0)

        contents = numpy.asarray(f["hpxpy"].bases[-1])
        assert (contents.min(), contents.max()) == (0.0, 497.0)

        contents = numpy.asarray(f["hprof"].bases[-1].bases[-1])
        assert (contents.min(), contents.max()) == (0.0, 3054.7299575805664)

        numpy.asarray(f["ntuple"])


def test_gohep_nosplit_file():
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-nosplit.root"))[
        "tree/evt"
    ] as branch:
        result = branch.array(library="np", entry_start=5, entry_stop=6)[0]
        assert result.member("Beg") == "beg-005"
        assert result.member("I16") == 5
        assert result.member("I32") == 5
        assert result.member("I64") == 5
        assert result.member("U16") == 5
        assert result.member("U32") == 5
        assert result.member("U64") == 5
        assert result.member("F32") == 5.0
        assert result.member("F64") == 5.0
        assert result.member("Str") == "evt-005"
        assert result.member("P3").member("Px") == 4
        assert result.member("P3").member("Py") == 5.0
        assert result.member("P3").member("Pz") == 4
        assert result.member("ArrayI16").tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
        assert result.member("ArrayU16").tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
        assert result.member("ArrayI32").tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
        assert result.member("ArrayU32").tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
        assert result.member("ArrayI64").tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
        assert result.member("ArrayU64").tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
        assert result.member("ArrayF32").tolist() == [
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ]
        assert result.member("ArrayF32").tolist() == [
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ]
        assert result.member("ArrayF64").tolist() == [
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ]
        assert result.member("ArrayF64").tolist() == [
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ]
        assert result.member("StdStr") == "std-005"
        assert result.member("SliceI16").tolist() == [5, 5, 5, 5, 5]
        assert result.member("SliceI32").tolist() == [5, 5, 5, 5, 5]
        assert result.member("SliceI64").tolist() == [5, 5, 5, 5, 5]
        assert result.member("SliceU16").tolist() == [5, 5, 5, 5, 5]
        assert result.member("SliceU32").tolist() == [5, 5, 5, 5, 5]
        assert result.member("SliceU64").tolist() == [5, 5, 5, 5, 5]
        assert result.member("StlVecI16").tolist() == [5, 5, 5, 5, 5]
        assert result.member("StlVecI32").tolist() == [5, 5, 5, 5, 5]
        assert result.member("StlVecI64").tolist() == [5, 5, 5, 5, 5]
        assert result.member("StlVecU16").tolist() == [5, 5, 5, 5, 5]
        assert result.member("StlVecU32").tolist() == [5, 5, 5, 5, 5]
        assert result.member("StlVecU64").tolist() == [5, 5, 5, 5, 5]
        assert result.member("StlVecF32").tolist() == [5.0, 5.0, 5.0, 5.0, 5.0]
        assert result.member("StlVecF64").tolist() == [5.0, 5.0, 5.0, 5.0, 5.0]
        assert result.member("StlVecStr").tolist() == [
            "vec-005",
            "vec-005",
            "vec-005",
            "vec-005",
            "vec-005",
        ]
        assert result.member("End") == "end-005"


def test_TLorentzVectors_show():
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        tree.show()


def test_TVector2():
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        result = tree["MET"].array(library="np", entry_stop=1)[0]
        assert (result.member("fX"), result.member("fY")) == (
            5.912771224975586,
            2.5636332035064697,
        )


def test_vector_TLorentzVector():
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        result = tree["muonp4"].array(library="np", entry_stop=1)[0]
        assert len(result) == 2
        assert (
            result[0].member("fE"),
            result[0].member("fP").member("fX"),
            result[0].member("fP").member("fY"),
            result[0].member("fP").member("fZ"),
        ) == (
            54.77949905395508,
            -52.89945602416992,
            -11.654671669006348,
            -8.16079330444336,
        )


def test_strided():
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        assert tree.file.class_named("TLorentzVector", "max").strided_interpretation(
            tree.file
        ).from_dtype == numpy.dtype(
            [
                ("@instance_version", numpy.dtype(">u2")),
                ("@num_bytes", numpy.dtype(">u4")),
                ("@fUniqueID", numpy.dtype(">u4")),
                ("@fBits", numpy.dtype(">u4")),
                ("@pidf", numpy.dtype(">u2")),
                ("fP/@instance_version", numpy.dtype(">u2")),
                ("fP/@num_bytes", numpy.dtype(">u4")),
                ("fP/@fUniqueID", numpy.dtype(">u4")),
                ("fP/@fBits", numpy.dtype(">u4")),
                ("fP/@pidf", numpy.dtype(">u2")),
                ("fP/fX", ">f8"),
                ("fP/fY", ">f8"),
                ("fP/fZ", ">f8"),
                ("fE", ">f8"),
            ]
        )


def test_read_strided_TVector2():
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        interp = tree.file.class_named("TVector2", "max").strided_interpretation(
            tree.file
        )

        assert tree["MET"].interpretation == interp

        result = tree["MET"].array(entry_stop=1, library="np")[0]
        assert (result.member("fX"), result.member("fY")) == (
            5.912771224975586,
            2.5636332035064697,
        )


def test_read_strided_TLorentzVector():
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        interp = tree.file.class_named("TLorentzVector", "max").strided_interpretation(
            tree.file
        )
        interp = uproot.interpretation.jagged.AsJagged(interp, header_bytes=10)

        assert tree["muonp4"].interpretation == interp

        result = tree["muonp4"].array(library="np", entry_stop=1)[0]
        assert len(result) == 2
        assert (
            result[0].member("fE"),
            result[0].member("fP").member("fX"),
            result[0].member("fP").member("fY"),
            result[0].member("fP").member("fZ"),
        ) == (
            54.77949905395508,
            -52.89945602416992,
            -11.654671669006348,
            -8.16079330444336,
        )


def test_strided_awkward():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        result = tree["MET"].array(library="ak")

        assert str(awkward.type(result)) == "2421 * TVector2[fX: float64, fY: float64]"

        assert awkward.to_list(result["fX"][:10]) == [
            5.912771224975586,
            24.76520347595215,
            -25.78508758544922,
            8.619895935058594,
            5.393138885498047,
            -3.7594752311706543,
            23.962148666381836,
            -57.533348083496094,
            42.416194915771484,
            -1.9144694805145264,
        ]
        assert awkward.to_list(result["fY"][:10]) == [
            2.5636332035064697,
            -16.349109649658203,
            16.237131118774414,
            -22.78654670715332,
            -1.3100523948669434,
            -19.417020797729492,
            -9.049156188964844,
            -20.48767852783203,
            -94.35086059570312,
            -23.96303367614746,
        ]


def test_jagged_strided_awkward():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        result = tree["muonp4"].array(library="ak")

        assert (
            str(awkward.type(result))
            == "2421 * var * TLorentzVector[fP: TVector3[fX: float64, "
            "fY: float64, fZ: float64], fE: float64]"
        )

        assert result[0, 0, "fE"] == 54.77949905395508
        assert result[0, 0, "fP", "fX"] == -52.89945602416992
        assert result[0, 0, "fP", "fY"] == -11.654671669006348
        assert result[0, 0, "fP", "fZ"] == -8.16079330444336


def test_jagged_awkward_1():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree/evt/SliceU64"
    ] as branch:
        assert branch.interpretation == AsJagged(AsDtype(">u8"), header_bytes=1)
        result = branch.array(library="ak", entry_stop=6)
        assert awkward.to_list(result) == [
            [],
            [1],
            [2, 2],
            [3, 3, 3],
            [4, 4, 4, 4],
            [5, 5, 5, 5, 5],
        ]


def test_jagged_awkward_2():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree/evt/StlVecF64"
    ] as branch:
        assert branch.interpretation == AsJagged(AsDtype(">f8"), header_bytes=10)
        result = branch.array(library="ak", entry_stop=6)
        assert awkward.to_list(result) == [
            [],
            [1],
            [2, 2],
            [3, 3, 3],
            [4, 4, 4, 4],
            [5, 5, 5, 5, 5],
        ]


def test_general_awkward_form():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-nosplit.root"))[
        "tree/evt"
    ] as branch:
        context = {
            "index_format": "i64",
            "header": False,
            "tobject_header": True,
            "breadcrumbs": (),
        }

        assert json.loads(
            branch.interpretation.awkward_form(branch.file, context).to_json()
        ) == json.loads(
            """{
    "class": "RecordArray",
    "fields": [
        "Beg",
        "I16",
        "I32",
        "I64",
        "U16",
        "U32",
        "U64",
        "F32",
        "F64",
        "Str",
        "P3",
        "ArrayI16",
        "ArrayI32",
        "ArrayI64",
        "ArrayU16",
        "ArrayU32",
        "ArrayU64",
        "ArrayF32",
        "ArrayF64",
        "N",
        "SliceI16",
        "SliceI32",
        "SliceI64",
        "SliceU16",
        "SliceU32",
        "SliceU64",
        "SliceF32",
        "SliceF64",
        "StdStr",
        "StlVecI16",
        "StlVecI32",
        "StlVecI64",
        "StlVecU16",
        "StlVecU32",
        "StlVecU64",
        "StlVecF32",
        "StlVecF64",
        "StlVecStr",
        "End"
    ],
    "contents": [
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint8",
                "inner_shape": [],
                "parameters": {
                    "__array__": "char"
                },
                "form_key": null
            },
            "parameters": {
                "__array__": "string"
            },
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "int16",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "int32",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "int64",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "uint16",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "uint32",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "uint64",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "float32",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "float64",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint8",
                "inner_shape": [],
                "parameters": {
                    "__array__": "char"
                },
                "form_key": null
            },
            "parameters": {
                "__array__": "string"
            },
            "form_key": null
        },
        {
            "class": "RecordArray",
            "fields": [
                "Px",
                "Py",
                "Pz"
            ],
            "contents": [
                {
                    "class": "NumpyArray",
                    "primitive": "int32",
                    "inner_shape": [],
                    "parameters": {},
                    "form_key": null
                },
                {
                    "class": "NumpyArray",
                    "primitive": "float64",
                    "inner_shape": [],
                    "parameters": {},
                    "form_key": null
                },
                {
                    "class": "NumpyArray",
                    "primitive": "int32",
                    "inner_shape": [],
                    "parameters": {},
                    "form_key": null
                }
            ],
            "parameters": {
                "__record__": "P3"
            },
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "int16",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "int32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "uint16",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "uint32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "uint64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "float32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "RegularArray",
            "size": 10,
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "NumpyArray",
            "primitive": "uint32",
            "inner_shape": [],
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int16",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint16",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint8",
                "inner_shape": [],
                "parameters": {
                    "__array__": "char"
                },
                "form_key": null
            },
            "parameters": {
                "__array__": "string"
            },
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int16",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint16",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float32",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "inner_shape": [],
                "parameters": {},
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "ListOffsetArray",
                "offsets": "i64",
                "content": {
                    "class": "NumpyArray",
                    "primitive": "uint8",
                    "inner_shape": [],
                    "parameters": {
                        "__array__": "char"
                    },
                    "form_key": null
                },
                "parameters": {
                    "__array__": "string"
                },
                "form_key": null
            },
            "parameters": {},
            "form_key": null
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "uint8",
                "inner_shape": [],
                "parameters": {
                    "__array__": "char"
                },
                "form_key": null
            },
            "parameters": {
                "__array__": "string"
            },
            "form_key": null
        }
    ],
    "parameters": {
        "__record__": "Event"
    },
    "form_key": null
}"""
        )


def test_jagged_awkward_3():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-fullsplit.root"))[
        "tree/evt/StlVecStr"
    ] as branch:
        assert awkward.to_list(branch.array(library="ak")[:6, :3]) == [
            [],
            ["vec-001"],
            ["vec-002", "vec-002"],
            ["vec-003", "vec-003", "vec-003"],
            ["vec-004", "vec-004", "vec-004"],
            ["vec-005", "vec-005", "vec-005"],
        ]


def test_awkward_string():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["string"].array(library="ak")) == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_awkward_tstring():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["tstring"].array(library="ak")) == [
            "one",
            "two",
            "three",
            "four",
            "five",
        ]


def test_awkward_vector_int32():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_int32"].array(library="ak")) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]


def test_awkward_vector_string():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_string"].array(library="ak")) == [
            ["one"],
            ["one", "two"],
            ["one", "two", "three"],
            ["one", "two", "three", "four"],
            ["one", "two", "three", "four", "five"],
        ]


def test_awkward_vector_tstring():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_tstring"].array(library="ak")) == [
            ["one"],
            ["one", "two"],
            ["one", "two", "three"],
            ["one", "two", "three", "four"],
            ["one", "two", "three", "four", "five"],
        ]


def test_awkward_vector_vector_int32():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_vector_int32"].array(library="ak")) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2], [1, 2, 3]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
        ]


def test_awkward_vector_vector_string():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        assert awkward.to_list(tree["vector_vector_string"].array(library="ak")) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "two", "three"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
                ["one", "two", "three", "four", "five"],
            ],
        ]


def test_awkward_map_int32_int16():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_int32_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]


def test_awkward_map_int32_vector_int16():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_int32_vector_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [[1]],
            [[1], [1, 2]],
            [[1], [1, 2], [1, 2, 3]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
        ]


def test_awkward_map_int32_vector_string():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(
            tree["map_int32_vector_string"].array(library="ak")
        )
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two"], ["one", "two", "three"]],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
            ],
            [
                ["one"],
                ["one", "two"],
                ["one", "two", "three"],
                ["one", "two", "three", "four"],
                ["one", "two", "three", "four", "five"],
            ],
        ]


def test_awkward_map_string_int16():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_int16"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            [1],
            [1, 2],
            [1, 3, 2],
            [4, 1, 3, 2],
            [5, 4, 1, 3, 2],
        ]


def test_awkward_map_string_vector_string():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(
            tree["map_string_vector_string"].array(library="ak")
        )
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            [["one"]],
            [["one"], ["one", "two"]],
            [["one"], ["one", "two", "three"], ["one", "two"]],
            [
                ["one", "two", "three", "four"],
                ["one"],
                ["one", "two", "three"],
                ["one", "two"],
            ],
            [
                ["one", "two", "three", "four", "five"],
                ["one", "two", "three", "four"],
                ["one"],
                ["one", "two", "three"],
                ["one", "two"],
            ],
        ]


def test_awkward_map_int32_vector_vector_int16():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(
            tree["map_int32_vector_vector_int16"].array(library="ak")
        )
        assert awkward.to_list(keys) == [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4],
            [1, 2, 3, 4, 5],
        ]
        assert awkward.to_list(values) == [
            [[[1]]],
            [[[1]], [[1], [1, 2]]],
            [[[1]], [[1], [1, 2]], [[1], [1, 2], [1, 2, 3]]],
            [
                [[1]],
                [[1], [1, 2]],
                [[1], [1, 2], [1, 2, 3]],
                [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
            ],
            [
                [[1]],
                [[1], [1, 2]],
                [[1], [1, 2], [1, 2, 3]],
                [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4]],
                [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]],
            ],
        ]


def test_awkward_map_string_string():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_string"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            ["ONE"],
            ["ONE", "TWO"],
            ["ONE", "THREE", "TWO"],
            ["FOUR", "ONE", "THREE", "TWO"],
            ["FIVE", "FOUR", "ONE", "THREE", "TWO"],
        ]


def test_awkward_map_string_tstring():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root"))[
        "tree"
    ] as tree:
        keys, values = awkward.unzip(tree["map_string_tstring"].array(library="ak"))
        assert awkward.to_list(keys) == [
            ["one"],
            ["one", "two"],
            ["one", "three", "two"],
            ["four", "one", "three", "two"],
            ["five", "four", "one", "three", "two"],
        ]
        assert awkward.to_list(values) == [
            ["ONE"],
            ["ONE", "TWO"],
            ["ONE", "THREE", "TWO"],
            ["FOUR", "ONE", "THREE", "TWO"],
            ["FIVE", "FOUR", "ONE", "THREE", "TWO"],
        ]


def test_awkward_map_int_struct():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-issue468.root"))[
        "Geant4Data/Geant4Data./Geant4Data.particles"
    ] as branch:
        assert (
            repr(branch.interpretation) == "AsObjects(AsMap(True, dtype('>i4'), "
            "Model_BDSOutputROOTGeant4Data_3a3a_ParticleInfo))"
        )
        result = branch.array(library="ak")
        assert (
            str(awkward.type(result))
            == '1 * var * tuple[[int32, struct[{name: string, charge: int32, mass: float64}, parameters={"__record__": "BDSOutputROOTGeant4Data::ParticleInfo"}]], parameters={"__array__": "sorted_map"}]'
        )
        assert awkward.to_list(result[0, "0"]) == [
            -1000020040,
            -1000020030,
            -1000010030,
            -1000010020,
            -2212,
            -2112,
            -321,
            -211,
            -16,
            -15,
            -14,
            -13,
            -12,
            -11,
            0,
            11,
            12,
            13,
            14,
            15,
            16,
            22,
            211,
            321,
            2112,
            2212,
            1000010020,
            1000010030,
            1000020030,
            1000020040,
        ]
        assert awkward.to_list(result[0, "1", "name"]) == [
            "anti_alpha",
            "anti_He3",
            "anti_triton",
            "anti_deuteron",
            "anti_proton",
            "anti_neutron",
            "kaon-",
            "pi-",
            "anti_nu_tau",
            "tau+",
            "anti_nu_mu",
            "mu+",
            "anti_nu_e",
            "e+",
            "geantino",
            "e-",
            "nu_e",
            "mu-",
            "nu_mu",
            "tau-",
            "nu_tau",
            "gamma",
            "pi+",
            "kaon+",
            "neutron",
            "proton",
            "deuteron",
            "triton",
            "He3",
            "alpha",
        ]
        assert awkward.to_list(result[0, "1", "charge"]) == [
            -2,
            -2,
            -1,
            -1,
            -1,
            0,
            -1,
            -1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            -1,
            0,
            -1,
            0,
            -1,
            0,
            0,
            1,
            1,
            0,
            1,
            1,
            1,
            2,
            2,
        ]
        assert awkward.to_list(result[0, "1", "mass"]) == [
            3.727379,
            2.808391,
            2.808921,
            1.875613,
            0.938272013,
            0.93956536,
            0.493677,
            0.1395701,
            0.0,
            1.77686,
            0.0,
            0.1056583715,
            0.0,
            0.00051099891,
            0.0,
            0.00051099891,
            0.0,
            0.1056583715,
            0.0,
            1.77686,
            0.0,
            0.0,
            0.1395701,
            0.493677,
            0.93956536,
            0.938272013,
            1.875613,
            2.808921,
            2.808391,
            3.727379,
        ]


def test_awkward_nosplit_file():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-small-evnt-tree-nosplit.root"))[
        "tree/evt"
    ] as branch:
        result = branch.array(library="ak", entry_stop=5)
        assert awkward.to_list(result["Beg"]) == [
            "beg-000",
            "beg-001",
            "beg-002",
            "beg-003",
            "beg-004",
        ]
        assert awkward.to_list(result["I16"]) == [0, 1, 2, 3, 4]
        assert awkward.to_list(result["F32"]) == [0.0, 1.0, 2.0, 3.0, 4.0]
        assert awkward.to_list(result["Str"]) == [
            "evt-000",
            "evt-001",
            "evt-002",
            "evt-003",
            "evt-004",
        ]
        assert awkward.to_list(result["P3", "Px"]) == [-1, 0, 1, 2, 3]
        assert awkward.to_list(result["P3", "Py"]) == [0.0, 1.0, 2.0, 3.0, 4.0]
        assert awkward.to_list(result["P3", "Pz"]) == [-1, 0, 1, 2, 3]
        assert awkward.to_list(result["ArrayI32"]) == [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            [3, 3, 3, 3, 3, 3, 3, 3, 3, 3],
            [4, 4, 4, 4, 4, 4, 4, 4, 4, 4],
        ]
        assert awkward.to_list(result["StdStr"]) == [
            "std-000",
            "std-001",
            "std-002",
            "std-003",
            "std-004",
        ]
        assert awkward.to_list(result["SliceI64"]) == [
            [],
            [1],
            [2, 2],
            [3, 3, 3],
            [4, 4, 4, 4],
        ]
        assert awkward.to_list(result["StlVecStr"]) == [
            [],
            ["vec-001"],
            ["vec-002", "vec-002"],
            ["vec-003", "vec-003", "vec-003"],
            ["vec-004", "vec-004", "vec-004", "vec-004"],
        ]
        assert awkward.to_list(result["End"]) == [
            "end-000",
            "end-001",
            "end-002",
            "end-003",
            "end-004",
        ]


def test_pandas_TVector2():
    pandas = pytest.importorskip("pandas")
    pytest.importorskip("awkward_pandas")
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        result = tree["MET"].array(library="pd")

        assert result.ak["fX"][0] == 5.912771224975586
        assert result.ak["fY"][0] == 2.5636332035064697


def test_pandas_vector_TLorentzVector():
    pandas = pytest.importorskip("pandas")
    pytest.importorskip("awkward_pandas")
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root"))[
        "events"
    ] as tree:
        result = tree["muonp4"].array(library="pd")

        assert result.ak["fP", "fX"][0].tolist() == [
            -52.89945602416992,
            37.7377815246582,
        ]
        assert result.ak["fP", "fY"][0].tolist() == [
            -11.654671669006348,
            0.6934735774993896,
        ]
        assert result.ak["fP", "fZ"][0].tolist() == [
            -8.16079330444336,
            -11.307581901550293,
        ]
        assert result.ak["fE"][0].tolist() == [
            54.77949905395508,
            39.401695251464844,
        ]


def test_map_string_TVector3():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-issue494.root"))[
        "Model/Model.scoringMeshTranslation"
    ] as branch:
        result = branch.array(library="ak")
        assert awkward.to_list(result["0"]) == [
            [
                "global_mesh",
                "mesh_foil1",
                "mesh_foil10",
                "mesh_foil11",
                "mesh_foil2",
                "mesh_foil3",
                "mesh_foil4",
                "mesh_foil5",
                "mesh_foil6",
                "mesh_foil7",
                "mesh_foil8",
                "mesh_foil9",
                "mesh_t1",
                "mesh_t2",
                "mesh_t3",
                "mesh_t4",
                "mesh_t5",
                "mesh_t6",
                "mesh_t7",
            ]
        ]
        assert awkward.to_list(result["1"]) == [
            [
                {"fX": 0.0, "fY": 0.0, "fZ": 0.074},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.048509515},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.09400956000000006},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.09500956300000007},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.049509518},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.050509521},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.079009536},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.08000953900000002},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.08100954200000002},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.08200954500000003},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.09200955400000005},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.09300955700000006},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.038000009},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.054000023999999994},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.06850002999999999},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.08550004800000004},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.09850006600000008},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.11300007200000009},
                {"fX": 0.0, "fY": 0.04, "fZ": 0.13600007800000008},
            ]
        ]


def test_gohep_output_file():
    awkward = pytest.importorskip("awkward")
    with uproot.open(skhep_testdata.data_path("uproot-issue413.root"))[
        "mytree"
    ] as tree:
        assert awkward.to_list(tree["I32"].array()) == [0, 1, 2, 3, 4]
        assert awkward.to_list(tree["F64"].array()) == [0.0, 1.0, 2.0, 3.0, 4.0]
        assert list(tree["Str"].array()) == [
            "evt-0",
            "evt-1",
            "evt-2",
            "evt-3",
            "evt-4",
        ]
        assert awkward.to_list(tree["ArrF64"].array()) == [
            [0.0, 1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0, 5.0],
            [2.0, 3.0, 4.0, 5.0, 6.0],
            [3.0, 4.0, 5.0, 6.0, 7.0],
            [4.0, 5.0, 6.0, 7.0, 8.0],
        ]
        assert awkward.to_list(tree["N"].array()) == [0, 1, 2, 3, 4]
        assert awkward.to_list(tree["SliF64"].array()) == [
            [],
            [1.0],
            [2.0, 3.0],
            [3.0, 4.0, 5.0],
            [4.0, 5.0, 6.0, 7.0],
        ]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys

import numpy
import pytest
import skhep_testdata

import uproot


def test_branchname():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        assert sample.arrays("i4", library="np")["i4"].tolist() == list(range(-15, 15))

        arrays = sample.arrays(["i4", "i8"], library="np")
        assert set(arrays.keys()) == {"i4", "i8"}
        assert arrays["i4"].tolist() == list(range(-15, 15))
        assert arrays["i8"].tolist() == list(range(-15, 15))

        arrays = sample.arrays(filter_name="/i[48]/", library="np")
        assert set(arrays.keys()) == {"i4", "i8"}
        assert arrays["i4"].tolist() == list(range(-15, 15))
        assert arrays["i8"].tolist() == list(range(-15, 15))

        arrays = sample.arrays(filter_name=["/i[12]/", "/i[48]/"], library="np")
        assert set(arrays.keys()) == {"i1", "i2", "i4", "i8"}
        assert arrays["i1"].tolist() == list(range(-15, 15))
        assert arrays["i2"].tolist() == list(range(-15, 15))
        assert arrays["i4"].tolist() == list(range(-15, 15))
        assert arrays["i8"].tolist() == list(range(-15, 15))

        arrays = sample.arrays(filter_name="i*", library="np")
        assert set(arrays.keys()) == {"i1", "i2", "i4", "i8"}
        assert arrays["i1"].tolist() == list(range(-15, 15))
        assert arrays["i2"].tolist() == list(range(-15, 15))
        assert arrays["i4"].tolist() == list(range(-15, 15))
        assert arrays["i8"].tolist() == list(range(-15, 15))

        arrays = sample.arrays(["i4", "i8"], filter_name="u*", library="np")
        assert set(arrays.keys()) == {"i4", "i8"}
        assert arrays["i4"].tolist() == list(range(-15, 15))
        assert arrays["i8"].tolist() == list(range(-15, 15))


def test_tuple_branchname():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        arrays = sample.arrays(["i4", "i8"], library="np", how=tuple)
        assert isinstance(arrays, tuple) and len(arrays) == 2
        assert arrays[0].tolist() == list(range(-15, 15))
        assert arrays[1].tolist() == list(range(-15, 15))

        arrays = sample.arrays(["i4", "i4"], library="np", how=tuple)
        assert isinstance(arrays, tuple) and len(arrays) == 2
        assert arrays[0].tolist() == list(range(-15, 15))
        assert arrays[1].tolist() == list(range(-15, 15))


def test_interpretation():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        assert sample["i2"].array(">u2", library="np").tolist() == list(
            range(65521, 65536)
        ) + list(range(0, 15))

        arrays = sample.arrays({"i1": ">u1", "i2": ">u2"}, library="np")
        assert set(arrays.keys()) == {"i1", "i2"}
        assert arrays["i1"].tolist() == list(range(241, 256)) + list(range(0, 15))
        assert arrays["i2"].tolist() == list(range(65521, 65536)) + list(range(0, 15))

        arrays = sample.arrays([("i1", ">u1"), ("i2", ">u2")], library="np", how=tuple)
        assert isinstance(arrays, tuple) and len(arrays) == 2
        assert arrays[0].tolist() == list(range(241, 256)) + list(range(0, 15))
        assert arrays[1].tolist() == list(range(65521, 65536)) + list(range(0, 15))

        arrays = sample.arrays({"i1": ">u1", "i2": None}, library="np")
        assert set(arrays.keys()) == {"i1", "i2"}
        assert arrays["i1"].tolist() == list(range(241, 256)) + list(range(0, 15))
        assert arrays["i2"].tolist() == list(range(-15, 15))

        arrays = sample.arrays([("i1", ">u1"), ("i2", None)], library="np", how=tuple)
        assert isinstance(arrays, tuple) and len(arrays) == 2
        assert arrays[0].tolist() == list(range(241, 256)) + list(range(0, 15))
        assert arrays[1].tolist() == list(range(-15, 15))

        with pytest.raises(ValueError):
            sample.arrays([("i1", ">u1"), ("i1", None)], library="np", how=tuple)


def test_compute():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        assert sample.arrays("i4 + 100", library="np")["i4 + 100"].tolist() == list(
            range(85, 115)
        )

        arrays = sample.arrays(["i4 + 100", "i8 + 100"], library="np")
        assert set(arrays.keys()) == {"i4 + 100", "i8 + 100"}
        assert arrays["i4 + 100"].tolist() == list(range(85, 115))
        assert arrays["i8 + 100"].tolist() == list(range(85, 115))

        arrays = sample.arrays(["i4 + 100", "i4 + 200"], library="np")
        assert set(arrays.keys()) == {"i4 + 100", "i4 + 200"}
        assert arrays["i4 + 100"].tolist() == list(range(85, 115))
        assert arrays["i4 + 200"].tolist() == list(range(185, 215))

        arrays = sample.arrays(["i4 + 100", "i4 + 100"], library="np")
        assert set(arrays.keys()) == {"i4 + 100"}
        assert arrays["i4 + 100"].tolist() == list(range(85, 115))

        arrays = sample.arrays(["i4 + 100", "i4 + 100"], library="np", how=tuple)
        assert isinstance(arrays, tuple) and len(arrays) == 2
        assert arrays[0].tolist() == list(range(85, 115))
        assert arrays[1].tolist() == list(range(85, 115))


def test_cut():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        assert sample.arrays("i4 + 100", cut="i4 > 0", library="np")[
            "i4 + 100"
        ].tolist() == list(range(101, 115))

        arrays = sample.arrays(["i4 + 100", "i8 + 100"], cut="i4 > 0", library="np")
        assert set(arrays.keys()) == {"i4 + 100", "i8 + 100"}
        assert arrays["i4 + 100"].tolist() == list(range(101, 115))
        assert arrays["i8 + 100"].tolist() == list(range(101, 115))

        arrays = sample.arrays(["i4", "i8"], cut="i4 > 0", library="np")
        assert set(arrays.keys()) == {"i4", "i8"}
        assert arrays["i4"].tolist() == list(range(1, 15))
        assert arrays["i8"].tolist() == list(range(1, 15))


def test_aliases():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        assert sample.arrays(
            "whatever", aliases={"whatever": "i4 + 100"}, library="np"
        )["whatever"].tolist() == list(range(85, 115))

        arrays = sample.arrays(
            ["one", "two"], aliases={"one": "i4 + 100", "two": "i8 + 100"}, library="np"
        )
        assert set(arrays.keys()) == {"one", "two"}
        assert arrays["one"].tolist() == list(range(85, 115))
        assert arrays["two"].tolist() == list(range(85, 115))

        arrays = sample.arrays(
            ["one", "two"], aliases={"one": "i4 + 100", "two": "one"}, library="np"
        )
        assert set(arrays.keys()) == {"one", "two"}
        assert arrays["one"].tolist() == list(range(85, 115))
        assert arrays["two"].tolist() == list(range(85, 115))

        with pytest.raises(ValueError):
            sample.arrays(
                ["one", "two"], aliases={"one": "two", "two": "one"}, library="np"
            )

        arrays = sample.arrays(
            ["one", "two"],
            cut="one > 100",
            aliases={"one": "i4 + 100", "two": "i8 + 100"},
            library="np",
        )
        assert set(arrays.keys()) == {"one", "two"}
        assert arrays["one"].tolist() == list(range(101, 115))
        assert arrays["two"].tolist() == list(range(101, 115))

        arrays = sample.arrays(
            ["i4"],
            cut="one > 100",
            aliases={"one": "i4 + 100", "two": "i8 + 100"},
            library="np",
        )
        assert set(arrays.keys()) == {"i4"}
        assert arrays["i4"].tolist() == list(range(1, 15))


def test_jagged():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    )["sample"] as sample:
        assert [x.tolist() for x in sample.arrays("Ai4", library="np")["Ai4"]] == [
            [],
            [-15],
            [-15, -13],
            [-15, -13, -11],
            [-15, -13, -11, -9],
            [],
            [-10],
            [-10, -8],
            [-10, -8, -6],
            [-10, -8, -6, -4],
            [],
            [-5],
            [-5, -3],
            [-5, -3, -1],
            [-5, -3, -1, 1],
            [],
            [0],
            [0, 2],
            [0, 2, 4],
            [0, 2, 4, 6],
            [],
            [5],
            [5, 7],
            [5, 7, 9],
            [5, 7, 9, 11],
            [],
            [10],
            [10, 12],
            [10, 12, 14],
            [10, 12, 14, 16],
        ]



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import sys

import numpy
import pytest
import skhep_testdata

import uproot


def tobytes(x):
    if hasattr(x, "tobytes"):
        return x.tobytes()
    else:
        return x.tostring()


def test_classname_encoding(tmpdir):
    assert (
        uproot.model.classname_encode("namespace::some.deep<templated, thing>", 12)
        == "Model_namespace_3a3a_some_2e_deep_3c_templated_2c20_thing_3e__v12"
    )
    assert (
        uproot.model.classname_encode("namespace::some.deep<templated, thing>")
        == "Model_namespace_3a3a_some_2e_deep_3c_templated_2c20_thing_3e_"
    )

    assert uproot.model.classname_decode(
        "Model_namespace_3a3a_some_2e_deep_3c_templated_2c20_thing_3e__v12"
    ) == ("namespace::some.deep<templated, thing>", 12)
    assert uproot.model.classname_decode(
        "Model_namespace_3a3a_some_2e_deep_3c_templated_2c20_thing_3e_"
    ) == ("namespace::some.deep<templated, thing>", None)


def test_file_header():
    filename = skhep_testdata.data_path("uproot-Zmumu.root")
    file = uproot.reading.ReadOnlyFile(filename)
    assert repr(file.compression) == "ZLIB(4)"
    assert not file.is_64bit
    assert file.fNbytesInfo == 4447
    assert file.hex_uuid == "944b77d0-98ab-11e7-a769-0100007fbeef"



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest
import skhep_testdata

import uproot
import uproot._util
import uproot.writing

ROOT = pytest.importorskip("ROOT")


def test_subdirs(tmp_path):
    filename = os.path.join(tmp_path, "testy.root")

    with uproot.writing.recreate(filename) as root_directory:
        subdir1 = root_directory.mkdir("wowzers")
        subdir2 = root_directory.mkdir("yikes")

    assert uproot.open(filename).classnames() == {
        "wowzers;1": "TDirectory",
        "yikes;1": "TDirectory",
    }

    f = ROOT.TFile(filename, "update")
    x = ROOT.TObjString("hello")
    x.Write()
    f.cd("wowzers")
    y = ROOT.TObjString("there")
    y.Write()
    f.cd("/yikes")
    z = ROOT.TObjString("you")
    z.Write()
    f.Close()

    assert uproot.open(filename).classnames() == {
        "wowzers;1": "TDirectory",
        "yikes;1": "TDirectory",
        "hello;1": "TObjString",
        "wowzers/there;1": "TObjString",
        "yikes/you;1": "TObjString",
    }

    g = ROOT.TFile(filename, "update")
    w = ROOT.TObjString("one_more")
    w.Write()
    g.Close()

    assert uproot.open(filename).classnames() == {
        "wowzers;1": "TDirectory",
        "yikes;1": "TDirectory",
        "hello;1": "TObjString",
        "one_more;1": "TObjString",
        "wowzers/there;1": "TObjString",
        "yikes/you;1": "TObjString",
    }


def test_subsubdir(tmp_path):
    filename = os.path.join(tmp_path, "testy.root")

    with uproot.writing.recreate(filename) as root_directory:
        subdir1 = root_directory.mkdir("wowzers")
        subdir2 = subdir1.mkdir("yikes")

    assert uproot.open(filename).classnames() == {
        "wowzers;1": "TDirectory",
        "wowzers/yikes;1": "TDirectory",
    }

    f = ROOT.TFile(filename, "update")
    x = ROOT.TObjString("hello")
    x.Write()
    f.cd("/wowzers")
    y = ROOT.TObjString("there")
    y.Write()
    f.cd("/wowzers/yikes")
    z = ROOT.TObjString("you")
    z.Write()
    f.Close()

    assert uproot.open(filename).classnames() == {
        "wowzers;1": "TDirectory",
        "wowzers/yikes;1": "TDirectory",
        "hello;1": "TObjString",
        "wowzers/there;1": "TObjString",
        "wowzers/yikes/you;1": "TObjString",
    }

    g = ROOT.TFile(filename, "update")
    g.cd("/wowzers/yikes")
    w = ROOT.TObjString("one_more")
    w.Write()
    g.Close()

    assert uproot.open(filename).classnames() == {
        "wowzers;1": "TDirectory",
        "wowzers/yikes;1": "TDirectory",
        "hello;1": "TObjString",
        "wowzers/there;1": "TObjString",
        "wowzers/yikes/you;1": "TObjString",
        "wowzers/yikes/one_more;1": "TObjString",
    }


def test_little_datime_functions():
    assert (
        uproot._util.datetime_to_code(uproot._util.code_to_datetime(1762860281))
        == 1762860281
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines a versionless model for ``ROOT::Experimental::RNTuple``.
"""

import queue
import struct

import awkward as ak
import numpy

import uproot

# https://github.com/root-project/root/blob/e9fa243af91217e9b108d828009c81ccba7666b5/tree/ntuple/v7/inc/ROOT/RMiniFile.hxx#L65
_rntuple_format1 = struct.Struct(">iIIQIIQIIQ")

# https://github.com/jblomer/root/blob/ntuple-binary-format-v1/tree/ntuple/v7/doc/specifications.md#envelopes
_rntuple_frame_format = struct.Struct("<HH")
_rntuple_feature_flag_format = struct.Struct("<Q")
_rntuple_num_bytes_fields = struct.Struct("<II")


def _renamemeA(chunk, cursor, context):
    env_version, min_version = cursor.fields(chunk, _rntuple_frame_format, context)
    return {"env_version": env_version, "min_version": min_version}


class Model_ROOT_3a3a_Experimental_3a3a_RNTuple(uproot.model.Model):
    """
    A versionless :doc:`uproot.model.Model` for ``ROOT::Experimental::RNTuple``.
    """

    @property
    def _keys(self):
        keys = []
        field_records = self.header.field_records
        for i, fr in enumerate(field_records):
            if fr.parent_field_id == i:
                keys.append(fr.field_name)
        return keys

    def keys(self):
        return self._keys

    def read_members(self, chunk, cursor, context, file):
        if self.is_memberwise:
            raise NotImplementedError(
                """memberwise serialization of {}
in file {}""".format(
                    type(self).__name__, self.file.file_path
                )
            )

        (
            self._members["fCheckSum"],
            self._members["fVersion"],
            self._members["fSize"],
            self._members["fSeekHeader"],
            self._members["fNBytesHeader"],
            self._members["fLenHeader"],
            self._members["fSeekFooter"],
            self._members["fNBytesFooter"],
            self._members["fLenFooter"],
            self._members["fReserved"],
        ) = cursor.fields(chunk, _rntuple_format1, context)

        seek, nbytes = self._members["fSeekHeader"], self._members["fNBytesHeader"]
        header_range = (seek, seek + nbytes)

        seek, nbytes = self._members["fSeekFooter"], self._members["fNBytesFooter"]
        footer_range = (seek, seek + nbytes)

        notifications = queue.Queue()
        compressed_header_chunk, compressed_footer_chunk = file.source.chunks(
            [header_range, footer_range], notifications=notifications
        )

        if self._members["fNBytesHeader"] == self._members["fLenHeader"]:
            self._header_chunk = compressed_header_chunk
            self._header_cursor = uproot.source.cursor.Cursor(
                self._members["fSeekHeader"]
            )
        else:
            self._header_chunk = uproot.compression.decompress(
                compressed_header_chunk,
                uproot.source.cursor.Cursor(self._members["fSeekHeader"]),
                context,
                self._members["fNBytesHeader"],
                self._members["fLenHeader"],
            )
            self._header_cursor = uproot.source.cursor.Cursor(0)

        if self._members["fNBytesFooter"] == self._members["fLenFooter"]:
            self._footer_chunk = compressed_footer_chunk
            self._footer_cursor = uproot.source.cursor.Cursor(
                self._members["fSeekFooter"]
            )
        else:
            self._footer_chunk = uproot.compression.decompress(
                compressed_footer_chunk,
                uproot.source.cursor.Cursor(self._members["fSeekFooter"]),
                context,
                self._members["fNBytesFooter"],
                self._members["fLenFooter"],
            )
            self._footer_cursor = uproot.source.cursor.Cursor(0)

        self._header, self._footer = None, None

        self._field_names = None
        self._column_records = None

        # back link from column name to page inner list
        self.column_innerlist_dict = {}

    @property
    def header(self):
        if self._header is None:
            cursor = self._header_cursor.copy()
            context = {}

            h = HeaderReader().read(self._header_chunk, cursor, context)
            self._header = h

        return self._header

    @property
    def field_names(self):
        if self._field_names is None:
            self._field_names = [r.field_name for r in self.header.field_records]
        return self._field_names

    @property
    def column_records(self):
        return self.header.column_records

    @property
    def footer(self):
        if self._footer is None:
            cursor = self._footer_cursor.copy()
            context = {}

            f = FooterReader().read(self._footer_chunk, cursor, context)
            assert (
                f.header_crc32 == self.header.crc32
            ), f"crc32={self.header.crc32}, header_crc32={f.header_crc32}"
            self._footer = f

        return self._footer

    @property
    def cluster_summaries(self):
        return self.footer.cluster_summaries

    # FIXME
    @property
    def _length(self):
        return self.cluster_summaries[0].num_entries

    def __len__(self):
        return self._length

    def which_colgroup(self, ncol):
        return 0  # FIXME haven't seen file with column group yet

    def read_locator(self, loc, uncomp_size, cursor, context):
        chunk = self.file.source.chunk(loc.offset, loc.offset + loc.num_bytes)
        if loc.num_bytes < uncomp_size:
            decomp_chunk = uproot.compression.decompress(
                chunk, cursor, context, loc.num_bytes, uncomp_size, block_info=None
            )
            cursor.move_to(0)
        else:
            decomp_chunk = chunk
        return decomp_chunk

    @property
    def page_list_envelopes(self):
        context = {}

        if not self.column_innerlist_dict:
            for record in self.footer.cluster_group_records:
                link = record.page_list_link
                loc = link.locator
                cursor = uproot.source.cursor.Cursor(loc.offset)
                decomp_chunk = self.read_locator(
                    loc, link.env_uncomp_size, cursor, context
                )
                self.column_innerlist_dict = PageLink().read(
                    decomp_chunk, cursor, context
                )

        return self.column_innerlist_dict

    def base_col_form(self, cr, col_id, parameters=None):
        form_key = f"column-{col_id}"
        dtype_byte = cr.type
        if dtype_byte == uproot.const.rntuple_role_union:
            return form_key
        elif dtype_byte > uproot.const.rntuple_role_struct:
            dt_str = uproot.const.rntuple_col_num_to_dtype_dict[dtype_byte]
            if dt_str == "bit":
                dt_str = "bool"
            return ak.forms.NumpyForm(
                dt_str,
                form_key=form_key,
                parameters=parameters,
            )
        else:  # offset index column
            return form_key

    def col_form(self, field_id, type_name=None):
        # FIXME remove this ugly logic
        rel_crs = []
        rel_crs_idxs = []
        for i, cr in enumerate(self.header.column_records):
            if cr.field_id == field_id:
                rel_crs.append(cr)
                rel_crs_idxs.append(i)
            if cr.field_id > field_id:
                break
        if len(rel_crs) == 1:  # base case
            return self.base_col_form(rel_crs[0], rel_crs_idxs[0])
        elif type_name == "std::string":  # string field splits->2 in col records
            assert len(rel_crs_idxs) == 2
            cr_char = rel_crs[-1]
            assert cr_char.type == uproot.const.rntuple_col_type_to_num_dict["char"]
            inner = self.base_col_form(
                cr_char, rel_crs_idxs[-1], parameters={"__array__": "char"}
            )
            form_key = f"column-{rel_crs_idxs[0]}"
            return ak.forms.ListOffsetForm(
                "u32", inner, form_key=form_key, parameters={"__array__": "string"}
            )
        else:
            raise (RuntimeError(f"Missing special case: {field_id}"))

    def field_form(self, this_id, seen):
        field_records = self.header.field_records
        this_record = field_records[this_id]
        seen.append(this_id)
        structural_role = this_record.struct_role
        if structural_role == uproot.const.rntuple_role_leaf:
            # base case of recursion
            # n.b. the split may happen in column
            return self.col_form(this_id, this_record.type_name)
        elif structural_role == uproot.const.rntuple_role_vector:
            keyname = self.col_form(this_id)
            child_id = next(
                filter(
                    lambda i: field_records[i].parent_field_id == this_id,
                    range(this_id + 1, len(field_records)),
                )
            )
            inner = self.field_form(child_id, seen)
            return ak.forms.ListOffsetForm("u32", inner, form_key=keyname)
        elif structural_role == uproot.const.rntuple_role_struct:
            newids = []
            for i, fr in enumerate(field_records):
                if i not in seen and fr.parent_field_id == this_id:
                    newids.append(i)
            # go find N in the rest, N is the # of fields in struct
            recordlist = [self.field_form(i, seen) for i in newids]
            namelist = [field_records[i].field_name for i in newids]
            return ak.forms.RecordForm(recordlist, namelist, form_key="whatever")
        elif structural_role == uproot.const.rntuple_role_union:
            keyname = self.col_form(this_id)
            newids = []
            for i, fr in enumerate(field_records):
                if i not in seen and fr.parent_field_id == this_id:
                    newids.append(i)
            recordlist = [self.field_form(i, seen) for i in newids]
            return ak.forms.UnionForm("i8", "i64", recordlist, form_key=keyname)
        else:
            # everything should recurse above this branch
            raise AssertionError("this should be unreachable")

    def to_akform(self):
        field_records = self.header.field_records
        recordlist = []
        topnames = self.keys()
        seen = []
        for i in range(len(field_records)):
            if i not in seen:
                recordlist.append(self.field_form(i, seen))

        form = ak.forms.RecordForm(recordlist, topnames, form_key="toplevel")
        return form

    def pagelist(self, listdesc):
        local_cursor = listdesc.cursor.copy()
        pages = listdesc.reader.read(listdesc.chunk, local_cursor, listdesc.context)
        return pages

    def read_pagedesc(self, destination, desc, dtype_str, dtype):
        loc = desc.locator
        cursor = uproot.source.cursor.Cursor(loc.offset)
        context = {}
        # bool in RNTuple is always stored as bits
        isbit = dtype_str == "bit"
        len_divider = 8 if isbit else 1
        num_elements = len(destination)
        num_elements_toread = int(numpy.ceil(num_elements / len_divider))
        uncomp_size = num_elements_toread * dtype.itemsize
        decomp_chunk = self.read_locator(loc, uncomp_size, cursor, context)
        content = cursor.array(
            decomp_chunk, num_elements_toread, dtype, context, move=False
        )
        if isbit:
            content = (
                numpy.unpackbits(content.view(dtype=numpy.uint8))
                .reshape(-1, 8)[:, ::-1]
                .reshape(-1)
            )

        # needed to chop off extra bits incase we used `unpackbits`
        destination[:] = content[:num_elements]

    def read_col_pages(self, ncol, cluster_range):
        return numpy.concatenate(
            [self.read_col_page(ncol, i) for i in cluster_range], axis=0
        )

    def read_col_page(self, ncol, cluster_i):
        linklist = self.page_list_envelopes.pagelinklist[cluster_i]
        link = linklist[ncol]
        pagelist = self.pagelist(link)
        dtype_byte = self.column_records[ncol].type
        dtype_str = uproot.const.rntuple_col_num_to_dtype_dict[dtype_byte]
        if dtype_str == "bit":
            dtype = numpy.dtype("bool")
        else:
            dtype = numpy.dtype(dtype_str)

        # FIXME vector read
        # n.b. it's possible pagelist is empty
        if not pagelist:
            return numpy.empty(0, dtype)
        total_len = numpy.sum([desc.num_elements for desc in pagelist])
        res = numpy.empty(total_len, dtype)
        tracker = 0
        for page_desc in pagelist:
            n_elements = page_desc.num_elements
            tracker_end = tracker + n_elements
            self.read_pagedesc(res[tracker:tracker_end], page_desc, dtype_str, dtype)
            tracker = tracker_end

        if dtype_byte <= uproot.const.rntuple_col_type_to_num_dict["index32"]:
            res = numpy.insert(res, 0, 0)  # for offsets
        return res

    def arrays(
        self,
        filter_names="*",
        filter_typenames=None,
        entry_start=0,
        entry_stop=None,
        decompression_executor=None,
        array_cache=None,
    ):
        entry_stop = entry_stop or self._length

        clusters = self.cluster_summaries
        cluster_starts = numpy.array([c.num_first_entry for c in clusters])

        start_cluster_idx = (
            numpy.searchsorted(cluster_starts, entry_start, side="right") - 1
        )
        stop_cluster_idx = numpy.searchsorted(cluster_starts, entry_stop, side="right")
        cluster_num_entries = numpy.sum(
            [c.num_entries for c in clusters[start_cluster_idx:stop_cluster_idx]]
        )

        form = self.to_akform().select_columns(filter_names)
        # only read columns mentioned in the awkward form
        target_cols = []
        container_dict = {}
        _recursive_find(form, target_cols)
        for i, cr in enumerate(self.column_records):
            key = f"column-{i}"
            dtype_byte = cr.type
            if key in target_cols:
                content = self.read_col_pages(
                    i, range(start_cluster_idx, stop_cluster_idx)
                )
                if dtype_byte == uproot.const.rntuple_col_type_to_num_dict["switch"]:
                    kindex, tags = _split_switch_bits(content)
                    container_dict[f"{key}-index"] = kindex
                    container_dict[f"{key}-tags"] = tags
                else:
                    # don't distinguish data and offsets
                    container_dict[f"{key}-data"] = content
                    container_dict[f"{key}-offsets"] = content
        cluster_offset = cluster_starts[start_cluster_idx]
        entry_start -= cluster_offset
        entry_stop -= cluster_offset
        return ak.from_buffers(form, cluster_num_entries, container_dict)[
            entry_start:entry_stop
        ]


# Supporting function and classes
def _split_switch_bits(content):
    kindex = numpy.bitwise_and(content, numpy.int64(0x00000000000FFFFF))
    tags = (content >> 44).astype("int8") - 1
    return kindex, tags


def _recursive_find(form, res):
    if hasattr(form, "form_key"):
        res.append(form.form_key)
    if hasattr(form, "contents"):
        for c in form.contents:
            _recursive_find(c, res)
    if hasattr(form, "content"):
        if issubclass(type(form.content), ak.forms.Form):
            _recursive_find(form.content, res)


# https://github.com/jblomer/root/blob/ntuple-binary-format-v1/tree/ntuple/v7/doc/specifications.md#page-list-envelope
class PageDescription:
    def read(self, chunk, cursor, context):
        out = MetaData(type(self).__name__)
        out.num_elements = cursor.field(chunk, struct.Struct("<I"), context)
        out.locator = LocatorReader().read(chunk, cursor, context)
        return out


class InnerListLocator:
    def __init__(self, chunk, cursor, context, num_pages):
        self.chunk = chunk
        self.cursor = cursor
        self.context = context
        self.num_pages = num_pages
        self.reader = ListFrameReader(PageDescription())
        self._page_descs = None

    def __repr__(self):
        return (
            f"InnerListLocator({self.chunk}, {self.cursor}, num_pages={self.num_pages})"
        )


class PageLinkInner:
    _frame_header = struct.Struct("<ii")

    def read(self, chunk, cursor, context):
        local_cursor = cursor.copy()
        # delay reading inner list of page descriptions
        future_cursor = cursor.copy()
        num_bytes, num_pages = local_cursor.fields(chunk, self._frame_header, context)
        assert num_bytes < 0, f"num_bytes={num_bytes}"
        cursor.skip(-num_bytes)
        return InnerListLocator(chunk, future_cursor, context, num_pages)


class PageLink:
    def __init__(self):
        self.top_most_list = ListFrameReader(  # top-most list
            ListFrameReader(PageLinkInner())  # outer list (inner list)
        )

    def read(self, chunk, cursor, context):
        out = MetaData(type(self).__name__)
        out.env_header = _renamemeA(chunk, cursor, context)

        out.pagelinklist = self.top_most_list.read(chunk, cursor, context)
        out.crc32 = cursor.field(chunk, struct.Struct("<I"), context)
        return out


class LocatorReader:
    _locator_format = struct.Struct("<iQ")

    def read(self, chunk, cursor, context):
        out = MetaData("Locator")
        out.num_bytes, out.offset = cursor.fields(chunk, self._locator_format, context)
        return out


class EnvLinkReader:
    def read(self, chunk, cursor, context):
        out = MetaData("EnvLink")
        out.env_uncomp_size = cursor.field(chunk, struct.Struct("<I"), context)
        out.locator = LocatorReader().read(chunk, cursor, context)
        return out


class MetaData:
    def __init__(self, name, **kwargs):
        self.__dict__["_name"] = name
        self.__dict__["_fields"] = kwargs

    @property
    def name(self):
        return self.__dict__["_name"]

    def __repr__(self):
        kwargs = ", ".join(f"{k}={v!r}" for k, v in self.__dict__["_fields"].items())
        return f"MetaData({self.name!r}, {kwargs})"

    def __getattr__(self, name):
        if not name.startswith("_"):
            return self.__dict__["_fields"][name]
        else:
            return self.__dict__[name]

    def __setattr__(self, name, val):
        self.__dict__["_fields"][name] = val


class RecordFrameReader:
    _record_size_format = struct.Struct("<I")

    def __init__(self, payload):
        self.payload = payload

    def read(self, chunk, cursor, context):
        local_cursor = cursor.copy()
        num_bytes = local_cursor.field(chunk, self._record_size_format, context)
        cursor.skip(num_bytes)
        return self.payload.read(chunk, local_cursor, context)


class ListFrameReader:
    _frame_header = struct.Struct("<ii")

    def __init__(self, payload):
        self.payload = payload

    def read(self, chunk, cursor, context):
        local_cursor = cursor.copy()
        num_bytes, num_items = local_cursor.fields(chunk, self._frame_header, context)
        assert num_bytes < 0, f"num_bytes={num_bytes}"
        cursor.skip(-num_bytes)
        return [
            self.payload.read(chunk, local_cursor, context) for _ in range(num_items)
        ]


# https://github.com/jblomer/root/blob/ntuple-binary-format-v1/tree/ntuple/v7/doc/specifications.md#field-description
class FieldRecordReader:
    _field_description = struct.Struct("<IIIHH")

    def read(self, chunk, cursor, context):
        out = MetaData("FieldRecordFrame")
        (
            out.field_version,
            out.type_version,
            out.parent_field_id,
            out.struct_role,
            out.flags,
        ) = cursor.fields(chunk, self._field_description, context)

        out.field_name, out.type_name, out.type_alias, out.field_desc = (
            cursor.rntuple_string(chunk, context) for i in range(4)
        )
        return out


# https://github.com/jblomer/root/blob/ntuple-binary-format-v1/tree/ntuple/v7/doc/specifications.md#column-description
class ColumnRecordReader:
    _column_record_format = struct.Struct("<HHII")

    def read(self, chunk, cursor, context):
        out = MetaData("ColumnRecordFrame")
        out.type, out.nbits, out.field_id, out.flags = cursor.fields(
            chunk, self._column_record_format, context
        )
        return out


class AliasColumnReader:
    _alias_column_ = struct.Struct("<II")

    def read(self, chunk, cursor, context):
        out = MetaData("AliasColumn")

        out.physical_id, out.field_id = cursor.fields(
            chunk, self._alias_column, context
        )
        return out


class ExtraTypeInfoReader:
    _extra_type_info = struct.Struct("<III")

    def read(self, chunk, cursor, context):
        out = MetaData("ExtraTypeInfoReader")

        out.type_ver_from, out.type_ver_to, out.content_id = cursor.fields(
            chunk, self._extra_type_info, context
        )
        out.type_name = cursor.rntuple_string(chunk, context)
        return out


class HeaderReader:
    def __init__(self):
        self.list_field_record_frames = ListFrameReader(
            RecordFrameReader(FieldRecordReader())
        )
        self.list_column_record_frames = ListFrameReader(
            RecordFrameReader(ColumnRecordReader())
        )
        self.list_alias_column_frames = ListFrameReader(
            RecordFrameReader(AliasColumnReader())
        )
        self.list_extra_type_info_reader = ListFrameReader(
            RecordFrameReader(ExtraTypeInfoReader())
        )

    def read(self, chunk, cursor, context):
        out = MetaData("Header")
        out.env_header = _renamemeA(chunk, cursor, context)
        out.feature_flag = cursor.field(chunk, _rntuple_feature_flag_format, context)
        out.rc_tag = cursor.field(chunk, struct.Struct("I"), context)
        out.name, out.ntuple_description, out.writer_identifier = (
            cursor.rntuple_string(chunk, context) for _ in range(3)
        )

        out.field_records = self.list_field_record_frames.read(chunk, cursor, context)
        out.column_records = self.list_column_record_frames.read(chunk, cursor, context)
        out.alias_columns = self.list_alias_column_frames.read(chunk, cursor, context)
        out.extra_type_infos = self.list_extra_type_info_reader.read(
            chunk, cursor, context
        )
        out.crc32 = cursor.field(chunk, struct.Struct("<I"), context)

        return out


class ColumnGroupRecordReader:
    def read(self, chunk, cursor, context):
        out = MetaData("ClusterSummaryRecord")
        out.num_first_entry, out.num_entries = cursor.fields(
            chunk, self._cluster_summary_format, context
        )
        return out


class ClusterSummaryReader:
    _cluster_summary_format = struct.Struct("<QQ")

    def read(self, chunk, cursor, context):
        out = MetaData("ClusterSummaryRecord")
        out.num_first_entry, out.num_entries = cursor.fields(
            chunk, self._cluster_summary_format, context
        )
        return out


class ClusterGroupRecordReader:
    _cluster_group_format = struct.Struct("<I")

    def read(self, chunk, cursor, context):
        out = MetaData("ClusterGroupRecord")
        out.num_clusters = cursor.field(chunk, self._cluster_group_format, context)
        out.page_list_link = EnvLinkReader().read(chunk, cursor, context)
        return out


class FooterReader:
    def __init__(self):
        self.extension_header_links = ListFrameReader(EnvLinkReader())
        self.column_group_record_frames = ListFrameReader(
            RecordFrameReader(ColumnGroupRecordReader())
        )
        self.cluster_summary_frames = ListFrameReader(
            RecordFrameReader(ClusterSummaryReader())
        )
        self.cluster_group_record_frames = ListFrameReader(
            RecordFrameReader(ClusterGroupRecordReader())
        )
        self.meta_data_links = ListFrameReader(EnvLinkReader())

    def read(self, chunk, cursor, context):
        out = MetaData("Footer")
        out.env_header = _renamemeA(chunk, cursor, context)
        out.feature_flag = cursor.field(chunk, _rntuple_feature_flag_format, context)
        out.header_crc32 = cursor.field(chunk, struct.Struct("<I"), context)

        out.extension_links = self.extension_header_links.read(chunk, cursor, context)
        out.col_group_records = self.column_group_record_frames.read(
            chunk, cursor, context
        )
        out.cluster_summaries = self.cluster_summary_frames.read(chunk, cursor, context)
        out.cluster_group_records = self.cluster_group_record_frames.read(
            chunk, cursor, context
        )
        out.meta_block_links = self.meta_data_links.read(chunk, cursor, context)
        return out


uproot.classes[
    "ROOT::Experimental::RNTuple"
] = Model_ROOT_3a3a_Experimental_3a3a_RNTuple



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os
import platform
import queue
import sys
from io import StringIO

import numpy
import pytest

import uproot


def tobytes(x):
    if hasattr(x, "tobytes"):
        return x.tobytes()
    else:
        return x.tostring()


def test_file(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")

    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    for num_workers in [1, 2]:
        source = uproot.source.file.MultithreadedFileSource(
            filename, num_workers=num_workers
        )
        with source as tmp:
            notifications = queue.Queue()
            chunks = tmp.chunks(
                [(0, 6), (6, 10), (10, 13), (13, 20), (20, 25), (25, 30)],
                notifications,
            )
            assert [tobytes(chunk.raw_data) for chunk in chunks] == [
                b"******",
                b"    ",
                b"...",
                b"+++++++",
                b"!!!!!",
                b"@@@@@",
            ]

        assert source.num_bytes == 30


def test_file_fail(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")

    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    for num_workers in [1, 2]:
        with pytest.raises(Exception):
            uproot.source.file.MultithreadedFileSource(
                filename + "-does-not-exist", num_workers=num_workers
            )


def test_memmap(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")

    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    source = uproot.source.file.MemmapSource(filename, num_fallback_workers=1)
    with source as tmp:
        notifications = queue.Queue()
        chunks = tmp.chunks(
            [(0, 6), (6, 10), (10, 13), (13, 20), (20, 25), (25, 30)], notifications
        )
        assert [tobytes(chunk.raw_data) for chunk in chunks] == [
            b"******",
            b"    ",
            b"...",
            b"+++++++",
            b"!!!!!",
            b"@@@@@",
        ]

        assert source.num_bytes == 30


def test_memmap_fail(tmpdir):
    filename = os.path.join(str(tmpdir), "tmp.raw")

    with open(filename, "wb") as tmp:
        tmp.write(b"******    ...+++++++!!!!!@@@@@")

    with pytest.raises(Exception):
        uproot.source.file.MultithreadedFileSource(filename + "-does-not-exist")


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
@pytest.mark.network
def test_http():
    source = uproot.source.http.HTTPSource(
        "https://example.com", timeout=10, num_fallback_workers=1
    )
    with source as tmp:
        notifications = queue.Queue()
        chunks = tmp.chunks([(0, 100), (50, 55), (200, 400)], notifications)
        one, two, three = (tobytes(chunk.raw_data) for chunk in chunks)
        assert len(one) == 100
        assert len(two) == 5
        assert len(three) == 200
    assert source.fallback is None

    source = uproot.source.http.MultithreadedHTTPSource(
        "https://example.com", num_workers=1, timeout=10
    )
    with source as tmp:
        notifications = queue.Queue()
        chunks = tmp.chunks([(0, 100), (50, 55), (200, 400)], notifications)
        assert [tobytes(x.raw_data) for x in chunks] == [one, two, three]


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
def colons_and_ports():
    assert uproot._util.file_object_path_split("https://example.com:443") == (
        "https://example.com:443",
        None,
    )
    assert uproot._util.file_object_path_split("https://example.com:443/something") == (
        "https://example.com:443/something",
        None,
    )
    assert uproot._util.file_object_path_split(
        "https://example.com:443/something:else"
    ) == ("https://example.com:443/something", "else")


@pytest.mark.skip(reason="RECHECK: example.com is flaky, too")
@pytest.mark.network
def test_http_port():
    source = uproot.source.http.HTTPSource(
        "https://example.com:443", timeout=10, num_fallback_workers=1
    )
    with source as tmp:
        notifications = queue.Queue()
        chunks = tmp.chunks([(0, 100), (50, 55), (200, 400)], notifications)
        one, two, three = (tobytes(chunk.raw_data) for chunk in chunks)
        assert len(one) == 100
        assert len(two) == 5
        assert len(three) == 200

    source = uproot.source.http.MultithreadedHTTPSource(
        "https://example.com:443", num_workers=1, timeout=10
    )
    with source as tmp:
        notifications = queue.Queue()
        chunks = tmp.chunks([(0, 100), (50, 55), (200, 400)], notifications)
        assert [tobytes(x.raw_data) for x in chunks] == [one, two, three]


@pytest.mark.network
def test_http_size():
    with uproot.source.http.HTTPSource(
        "https://scikit-hep.org/uproot3/examples/Zmumu.root",
        timeout=10,
        num_fallback_workers=1,
    ) as source:
        size1 = source.num_bytes

    with uproot.source.http.MultithreadedHTTPSource(
        "https://scikit-hep.org/uproot3/examples/Zmumu.root", num_workers=1, timeout=10
    ) as source:
        size2 = source.num_bytes

    assert size1 == size2


@pytest.mark.network
def test_http_size_port():
    with uproot.source.http.HTTPSource(
        "https://scikit-hep.org:443/uproot3/examples/Zmumu.root",
        timeout=10,
        num_fallback_workers=1,
    ) as source:
        size1 = source.num_bytes

    with uproot.source.http.MultithreadedHTTPSource(
        "https://scikit-hep.org:443/uproot3/examples/Zmumu.root",
        num_workers=1,
        timeout=10,
    ) as source:
        size2 = source.num_bytes

    assert size1 == size2


@pytest.mark.network
def test_http_fail():
    source = uproot.source.http.HTTPSource(
        "https://wonky.cern/does-not-exist", timeout=0.1, num_fallback_workers=1
    )
    with pytest.raises(Exception) as err:
        notifications = queue.Queue()
        chunks = source.chunks([(0, 100), (50, 55), (200, 400)], notifications)
        chunks[0].raw_data


@pytest.mark.network
def test_no_multipart():
    for num_workers in [1, 2]:
        with uproot.source.http.MultithreadedHTTPSource(
            "https://scikit-hep.org/uproot3/examples/Zmumu.root",
            num_workers=num_workers,
            timeout=10,
        ) as source:
            notifications = queue.Queue()
            chunks = source.chunks([(0, 100), (50, 55), (200, 400)], notifications)
            one, two, three = (tobytes(chunk.raw_data) for chunk in chunks)
            assert len(one) == 100
            assert len(two) == 5
            assert len(three) == 200
            assert one[:4] == b"root"


@pytest.mark.network
def test_no_multipart_fail():
    for num_workers in [1, 2]:
        source = uproot.source.http.MultithreadedHTTPSource(
            "https://wonky.cern/does-not-exist", num_workers=num_workers, timeout=0.1
        )
        with pytest.raises(Exception) as err:
            notifications = queue.Queue()
            chunks = source.chunks([(0, 100), (50, 55), (200, 400)], notifications)
            chunks[0].raw_data


@pytest.mark.network
def test_fallback():
    for num_workers in [1, 2]:
        with uproot.source.http.HTTPSource(
            "https://scikit-hep.org/uproot3/examples/Zmumu.root",
            timeout=10,
            num_fallback_workers=num_workers,
        ) as source:
            notifications = queue.Queue()
            chunks = source.chunks([(0, 100), (50, 55), (200, 400)], notifications)
            one, two, three = (tobytes(chunk.raw_data) for chunk in chunks)
            assert len(one) == 100
            assert len(two) == 5
            assert len(three) == 200
            assert one[:4] == b"root"


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.MultithreadedXRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        num_workers=1,
        timeout=20,
    ) as source:
        notifications = queue.Queue()
        chunks = source.chunks([(0, 100), (50, 55), (200, 400)], notifications)
        one, two, three = (tobytes(chunk.raw_data) for chunk in chunks)
        assert len(one) == 100
        assert len(two) == 5
        assert len(three) == 200
        assert one[:4] == b"root"


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_deadlock():
    pytest.importorskip("XRootD")
    # Attach this file to the "test_xrootd_deadlock" function so it leaks
    pytest.uproot_test_xrootd_deadlock_f = uproot.source.xrootd.XRootDResource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=20,
    )


@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_fail():
    pytest.importorskip("XRootD")
    with pytest.raises(Exception) as err:
        source = uproot.source.xrootd.MultithreadedXRootDSource(
            "root://wonky.cern/does-not-exist", num_workers=1, timeout=1
        )


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_vectorread():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.XRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=10,
        max_num_elements=None,
        num_workers=1,
    ) as source:
        notifications = queue.Queue()
        chunks = source.chunks([(0, 100), (50, 55), (200, 400)], notifications)
        one, two, three = (tobytes(chunk.raw_data) for chunk in chunks)
        assert len(one) == 100
        assert len(two) == 5
        assert len(three) == 200
        assert one[:4] == b"root"


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_vectorread_max_element_split():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.XRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=10,
        max_num_elements=None,
        num_workers=1,
    ) as source:
        notifications = queue.Queue()
        max_element_size = 2097136
        chunks = source.chunks([(0, max_element_size + 1)], notifications)
        (one,) = (tobytes(chunk.raw_data) for chunk in chunks)
        assert len(one) == max_element_size + 1


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_vectorread_max_element_split_consistency():
    pytest.importorskip("XRootD")
    filename = "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root"

    def get_chunk(Source, **kwargs):
        with Source(filename, **kwargs) as source:
            notifications = queue.Queue()
            max_element_size = 2097136
            chunks = source.chunks([(0, max_element_size + 1)], notifications)
            (one,) = (tobytes(chunk.raw_data) for chunk in chunks)
            return one

    chunk1 = get_chunk(
        uproot.source.xrootd.XRootDSource,
        timeout=10,
        max_num_elements=None,
        num_workers=1,
    )
    chunk2 = get_chunk(
        uproot.source.xrootd.MultithreadedXRootDSource, timeout=10, num_workers=1
    )
    assert chunk1 == chunk2


@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_vectorread_fail():
    pytest.importorskip("XRootD")
    with pytest.raises(Exception) as err:
        source = uproot.source.xrootd.XRootDSource(
            "root://wonky.cern/does-not-exist",
            timeout=1,
            max_num_elements=None,
            num_workers=1,
        )


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_size():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.XRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=10,
        max_num_elements=None,
        num_workers=1,
    ) as source:
        size1 = source.num_bytes

    pytest.importorskip("XRootD")
    with uproot.source.xrootd.MultithreadedXRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=10,
        num_workers=1,
    ) as source:
        size2 = source.num_bytes

    assert size1 == size2
    assert size1 == 3469136394


@pytest.mark.skip(
    reason="RECHECK: Run2012B_DoubleMuParked.root is super-flaky right now"
)
@pytest.mark.network
@pytest.mark.xrootd
def test_xrootd_numpy_int():
    pytest.importorskip("XRootD")
    with uproot.source.xrootd.XRootDSource(
        "root://eospublic.cern.ch//eos/root-eos/cms_opendata_2012_nanoaod/Run2012B_DoubleMuParked.root",
        timeout=10,
        max_num_elements=None,
        num_workers=1,
    ) as source:
        chunk = source.chunk(numpy.int64(0), numpy.int64(100))
        assert len(chunk.raw_data) == 100


def test_cursor_debug():
    data = numpy.concatenate(
        [
            numpy.array([123, 123, 123], "u1"),
            numpy.array(
                [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 101, 202, 303], ">f4"
            ).view("u1"),
            numpy.array([123, 123], "u1"),
        ]
    )
    future = uproot.source.futures.TrivialFuture(data)

    chunk = uproot.source.chunk.Chunk(None, 0, len(data), future)
    cursor = uproot.source.cursor.Cursor(0)

    output = StringIO()
    cursor.debug(chunk, offset=3, dtype=">f4", stream=output)
    assert (
        output.getvalue()
        == """--+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
123 123 123  63 140 204 205  64  12 204 205  64  83  51  51  64 140 204 205  64
  {   {   {   ? --- --- ---   @ --- --- ---   @   S   3   3   @ --- --- ---   @
                        1.1             2.2             3.3             4.4
    --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
    176   0   0  64 211  51  51  64 246 102 102  65  12 204 205  65  30 102 102  66
    --- --- ---   @ ---   3   3   @ ---   f   f   A --- --- ---   A ---   f   f   B
            5.5             6.6             7.7             8.8             9.9
    --+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
    202   0   0  67  74   0   0  67 151 128   0 123 123
    --- --- ---   C   J --- ---   C --- --- ---   {   {
          101.0           202.0           303.0
"""
    )

    output = StringIO()
    cursor.debug(chunk, stream=output)
    assert (
        output.getvalue()
        == """--+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
123 123 123  63 140 204 205  64  12 204 205  64  83  51  51  64 140 204 205  64
  {   {   {   ? --- --- ---   @ --- --- ---   @   S   3   3   @ --- --- ---   @
--+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
176   0   0  64 211  51  51  64 246 102 102  65  12 204 205  65  30 102 102  66
--- --- ---   @ ---   3   3   @ ---   f   f   A --- --- ---   A ---   f   f   B
--+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-
202   0   0  67  74   0   0  67 151 128   0 123 123
--- --- ---   C   J --- ---   C --- --- ---   {   {
"""
    )
